/* input.h -- definitions for es lexical analyzer ($Revision: 1.1.1.1 $) */
#ifndef __es__input_h__
#define __es__input_h__

#ifdef __cplusplus
extern "C" { 
#endif

#define	MAXUNGET	2		/* maximum 2 character pushback */

typedef struct Input Input;
struct Input {
  int ( *get )( Input *self );
  int ( *fill )( Input *self ), ( *rfill )( Input *self );
  void ( *cleanup )( Input *self );
  Input *        prev;
  const char *   name;
  unsigned char *buf, *bufend, *bufbegin, *rbuf;
  size_t         buflen;
  int            unget[ MAXUNGET ];
  int            ungot;
  int            lineno;
  int            fd;
  int            runflags;
};

#define GETC() ( *input->get )( input )
#define UNGETC( c ) unget( input, c )

/* eval_* flags are also understood as runflags */
#define RUN_INTERACTIVE          4      /* -i or $0[0] = '-' */
#define RUN_NOEXEC               8      /* -n */
#define RUN_ECHOINPUT           16      /* -v */
#define RUN_PRINTCMDS           32      /* -x */
#define RUN_LISPTREES           64      /* -L and defined(LISPTREES) */
#define RUN_INTERRUPT          128      /* ctrl-c on input */

/* input.c */
extern Boolean heredoc_input;
extern Boolean reset_terminal;
extern Input *input;
#define MAX_PROMPT_COUNT 16
extern Boolean is_prompt2; /* display secondary prompt */

void yyerror( char *s );
void sethistory( char *file );
void setevalstatus( char *status );
int keybind( char **args, int argc );
void unget( Input *in, int c );
void initgetenv( void );
Tree *parse( char **pr );
void resetparser( void );
List *runinput( Input *in, int runflags );
List *runfd( int fd, const char *name, int flags );
List *runstring( const char *str, const char *name, int flags );
Tree *parseinput( Input *in );
Tree *parsestring( const char *str );
Boolean isinteractive( void );
void initinput( void );

/* token.c */
extern const char non_word[];
extern const char dollar_non_word[];
int yylex( void );
void inityy( void );
void print_prompt2( void );

/* parse.y */
extern Tree *parsetree;
int yyparse( void );
void initparse( void );

/* heredoc.c */
void emptyherequeue( void );

#ifdef __cplusplus
}
#endif

#endif
